/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1996 David S. Miller (dm@engr.sgi.com)
 * Copyright (C) 1999 Andrew R. Baker (andrewb@uab.edu)
 * Copyright (C) 2001 Florian Lohoff (flo@rfc822.org)
 * Copyright (C) 2003 Ralf Baechle (ralf@linux-mips.org)
 * Copyright (C) 2004 Peter Fuerst (pf@net.alphadv.de) - IP28
 *
 * (In all truth, Jed Schimmel wrote all this code.)
 */
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/types.h>
#include <linux/mm.h>
#include <linux/blk.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/spinlock.h>

#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/sgialib.h>
#include <asm/sgi/sgi.h>
#include <asm/sgi/mc.h>
#include <asm/sgi/hpc3.h>
#include <asm/sgi/ip22.h>
#include <asm/irq.h>
#include <asm/io.h>

#include "scsi.h"
#include "hosts.h"
#include "wd33c93.h"
#include "sgiwd93.h"

#include <linux/stat.h>

#if 0
#define DPRINTK(args...)	printk(args)
#else
#define DPRINTK(args...)
#endif

#if defined(CONFIG_SGI_IP28) || defined(CONFIG_SGI_IP26)
/* see SGI documents: 007-0911-210, Chapter I.1. and man ip26_ucmem(D3) */
# define DMA_NONCOHERENT_CODE(code...) code
# define IP28_IP26_CODE(code...) code	/* avoid uncached writes */
#else
# if defined(CONFIG_DMA_NONCOHERENT) || defined(CONFIG_NONCOHERENT_IO)
#  define DMA_NONCOHERENT_CODE(code...) code
# else
#  define DMA_NONCOHERENT_CODE(code...)
# endif
# define IP28_IP26_CODE(code...)
#endif

/* There really should be common macros for this in addrspace.h :-( */
#ifdef CONFIG_XKPHYS_KERNEL
# define WD93CAC(p)   PHYS_TO_XKSEG_CACHED(WD93PHYS(p))
# define WD93UNCAC(p) PHYS_TO_XKSEG_UNCACHED(WD93PHYS(p))
# define WD93PHYS(p)  XKPHYS_TO_PHYS((unsigned long)(p))
#else
# define WD93CAC(p)   CKSEG0ADDR(p)
# define WD93UNCAC(p) CKSEG1ADDR(p)
# define WD93PHYS(p)  CPHYSADDR(p)
#endif

/* Alignment to the largest IP2X cacheblock. */
#define CACHLIGNDN(x)  ((x) & ~0x7f)
#define CACHLIGNUP(x)  CACHLIGNDN((x)+0x7f)

#define HDATA(ptr) ((struct ip22_hostdata *)((ptr)->hostdata))

struct ip22_hostdata {
	struct WD33C93_hostdata wh;
	struct hpc_data {
		dma_addr_t      dma;
		void            * cpu;
		/*
		 * A cacheline-aligned intermediate dma read buffer for all
		 * the small, not cacheline-aligned Scsi_Cmnd input buffers.
		 */
		char        (*buf)[PAGE_SIZE];
		struct {
			char     *scp;
			char     *rbp;
			unsigned len;
		} rb;
	} hd;
};

struct hpc_chunk {
	struct hpc_dma_desc desc;
	u32 _padding;	/* align to quadword boundary */
};

static dma_addr_t sgiwd93_dma_map(Scsi_Cmnd*);
static void sgiwd93_dma_unmap(Scsi_Cmnd*);

struct Scsi_Host *sgiwd93_host;
struct Scsi_Host *sgiwd93_host1;

/* Wuff wuff, wuff, wd33c93.c, wuff wuff, object oriented, bow wow. */
static inline void write_wd33c93_count(const wd33c93_regs regs,
                                      unsigned long value)
{
	*regs.SASR = WD_TRANSFER_COUNT_MSB;
	mb();
	*regs.SCMD = ((value >> 16) & 0xff);
	*regs.SCMD = ((value >>  8) & 0xff);
	*regs.SCMD = ((value >>  0) & 0xff);
	mb();
}

static inline unsigned long read_wd33c93_count(const wd33c93_regs regs)
{
	unsigned long value;

	*regs.SASR = WD_TRANSFER_COUNT_MSB;
	mb();
	value =  ((*regs.SCMD & 0xff) << 16);
	value |= ((*regs.SCMD & 0xff) <<  8);
	value |= ((*regs.SCMD & 0xff) <<  0);
	mb();
	return value;
}

static void sgiwd93_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;

	spin_lock_irqsave(&io_request_lock, flags);
	wd33c93_intr((struct Scsi_Host *) dev_id);
	spin_unlock_irqrestore(&io_request_lock, flags);
}

static inline int fill_hpc_entries(struct hpc_chunk *hcp, Scsi_Cmnd *cmd)
{
	unsigned long len = cmd->SCp.this_residual;
	dma_addr_t physaddr;
	unsigned long count;
	IP28_IP26_CODE( int nc; struct hpc_chunk *hcpstart; )

	physaddr = sgiwd93_dma_map(cmd);
	if (!physaddr)
		return 1;

	IP28_IP26_CODE(
		hcpstart = hcp = (typeof(hcp)) WD93CAC(hcp);
		nc = 1 + (len + 8191) / 8192;
		nc = CACHLIGNUP(sizeof(*hcp)*nc);
		/* dma_cache_inv((unsigned long) hcp, nc); */
	)
	while (len) {
		/*
		 * even cntinfo could be up to 16383, without
		 * magic only 8192 works correctly
		 * (even much less than 8192 may fail, since
		 * pbuf must not cross any page boundaries !)
		 */
		count = len > 8192 ? 8192 : len;
		hcp->desc.pbuf = physaddr;
		hcp->desc.cntinfo = count;
		hcp++;
		len -= count;
		physaddr += count;
	}

	/*
	 * To make sure, if we trip an HPC bug, that we transfer every single
	 * byte, we tag on an extra zero length dma descriptor at the end of
	 * the chain.
	 */
	hcp->desc.pbuf = 0;
	hcp->desc.cntinfo = HPCDMA_EOX;

	IP28_IP26_CODE( dma_cache_wback_inv((unsigned long)hcpstart, nc); )
	return 0;
}

static int dma_setup(Scsi_Cmnd *cmd, int datainp)
{
	struct ip22_hostdata *hdata = HDATA(cmd->host);
	struct hpc3_scsiregs *hregs =
		(struct hpc3_scsiregs *) cmd->host->base;
	struct hpc_chunk *hcp = (struct hpc_chunk *) hdata->hd.cpu;

	DPRINTK("dma_setup: datainp<%d> hcp<%p> ", datainp, hcp);

	hdata->wh.dma_dir = datainp;

	/*
	 * wd33c93 shouldn't pass us bogus dma_setups, but it does:-(  The
	 * other wd33c93 drivers deal with it the same way (which isn't that
	 * obvious).  IMHO a better fix would be, not to do these dma setups
	 * in the first place.
	 */
	if (cmd->SCp.ptr == NULL || cmd->SCp.this_residual == 0) {
		DPRINTK("\n");
		return 1;
	}
	if (fill_hpc_entries(hcp, cmd)) {
		DPRINTK("\n");
		return 1;
	}
	DPRINTK(" HPCGO\n");

	/* Start up the HPC. */
	hregs->ndptr = hdata->hd.dma;
	if (datainp)
		hregs->ctrl = HPC3_SCTRL_ACTIVE;
	else
		hregs->ctrl = HPC3_SCTRL_ACTIVE | HPC3_SCTRL_DIR;

	return 0;
}

static void dma_stop(struct Scsi_Host *instance, Scsi_Cmnd *SCpnt,
		     int status)
{
	struct ip22_hostdata *hdata = HDATA(instance);
	struct hpc3_scsiregs *hregs;

	if (!SCpnt)
		return; /* really ignore wd33c93_reset() ? */

	hregs = (struct hpc3_scsiregs *) SCpnt->host->base;

	DPRINTK("dma_stop: status<%d> ", status);

	/* First stop the HPC and flush it's FIFO. */
	if (hdata->wh.dma_dir) {
		hregs->ctrl |= HPC3_SCTRL_FLUSH;
		while (hregs->ctrl & HPC3_SCTRL_ACTIVE)
			barrier();
	}
	hregs->ctrl = 0;
	sgiwd93_dma_unmap(SCpnt);

	DPRINTK("\n");
}

void sgiwd93_reset(unsigned long base)
{
	struct hpc3_scsiregs *hregs = (struct hpc3_scsiregs *) base;

	hregs->ctrl = HPC3_SCTRL_CRESET;
	udelay(50);
	hregs->ctrl = 0;
}

static inline void init_hpc_chain(struct hpc_data *hd)
{
	struct hpc_chunk *hcp = (struct hpc_chunk *) hd->cpu;
	struct hpc_chunk *dma = (struct hpc_chunk *) hd->dma;
	unsigned long start, end;

	IP28_IP26_CODE( hcp = (typeof(hcp)) WD93CAC(hcp); )
	start = (unsigned long) hcp;
	end = start + PAGE_SIZE;
	while (start < end) {
		hcp->desc.pnext = (u32) (dma + 1);
		hcp->desc.cntinfo = HPCDMA_EOX;
		hcp++; dma++;
		start += sizeof(struct hpc_chunk);
	};
	hcp--;
	hcp->desc.pnext = hd->dma;

	IP28_IP26_CODE(
		/* Force flush to memory and purge cached references ! */
		dma_cache_wback_inv(WD93CAC(hd->cpu), PAGE_SIZE);
	)
	hd->cpu = (void *) WD93UNCAC(hd->cpu);
}

static struct Scsi_Host * __init sgiwd93_setup_scsi(
	Scsi_Host_Template *SGIblows, int unit, int irq,
	struct hpc3_scsiregs *hregs, unsigned char *wdregs)
{
	struct ip22_hostdata *hdata;
	struct Scsi_Host *host;
	wd33c93_regs regs;

	host = scsi_register(SGIblows, sizeof(struct ip22_hostdata));
	if (!host)
		return NULL;

	host->base = (unsigned long) hregs;
	host->irq = irq;

	hdata = HDATA(host);
	hdata->hd.cpu = pci_alloc_consistent(NULL, DMA_NONCOHERENT_CODE(2*)PAGE_SIZE,
	                                     &hdata->hd.dma);
	if (!hdata->hd.cpu) {
		printk(KERN_WARNING "sgiwd93: Could not allocate memory for "
		       "host %d buffer.\n", unit);
		goto out_unregister;
	}
	hdata->hd.rb.rbp = 0;
	hdata->hd.rb.scp = 0;
	hdata->hd.rb.len = 0;
	hdata->hd.buf = 0;
	DMA_NONCOHERENT_CODE(
		hdata->hd.buf = (void*) ((char*)hdata->hd.cpu + PAGE_SIZE);
	)
	init_hpc_chain(&hdata->hd);

	regs.SASR = wdregs + 3;
	regs.SCMD = wdregs + 7;

	wd33c93_init(host, regs, dma_setup, dma_stop, WD33C93_FS_16_20);

	hdata->wh.no_sync = 0;

	if (request_irq(irq, sgiwd93_intr, 0, "SGI WD93", (void *) host)) {
		printk(KERN_WARNING "sgiwd93: Could not register irq %d "
		       "for host %d.\n", irq, unit);
		goto out_free;
	}
	return host;

out_free:
	pci_free_consistent(NULL, DMA_NONCOHERENT_CODE(2*)PAGE_SIZE,
	                    hdata->hd.cpu, hdata->hd.dma);
	wd33c93_release();

out_unregister:
	scsi_unregister(host);

	return NULL;
}

int __init sgiwd93_detect(Scsi_Host_Template *SGIblows)
{
	int found = 0;

	SGIblows->proc_name = "SGIWD93";
	sgiwd93_host = sgiwd93_setup_scsi(SGIblows, 0, SGI_WD93_0_IRQ,
	                                  &hpc3c0->scsi_chan0,
	                                  (unsigned char *)hpc3c0->scsi0_ext);
	if (sgiwd93_host)
		found++;

	/* Set up second controller on the Indigo2 */
	if (ip22_is_fullhouse()) {
		sgiwd93_host1 = sgiwd93_setup_scsi(SGIblows, 1, SGI_WD93_1_IRQ,
		                          &hpc3c0->scsi_chan1,
		                          (unsigned char *)hpc3c0->scsi1_ext);
		if (sgiwd93_host1)
			found++;
	}

	return found;
}

static dma_addr_t sgiwd93_dma_map( Scsi_Cmnd *cmd )
{
	struct ip22_hostdata *hdata = HDATA(cmd->host);
	unsigned len = cmd->SCp.this_residual;
	void *addr = cmd->SCp.ptr;

	DMA_NONCOHERENT_CODE(
	if (hdata->hd.rb.rbp) {
		printk(KERN_ERR "sgiwd93_dma_map: %s: discarding stale input buffer!"
			" (%p)\n", hdata->wh.dma_dir ? "in":"out", hdata->hd.rb.scp);
		if (hdata->hd.rb.rbp != hdata->hd.buf)
			kfree(hdata->hd.rb.rbp);
	}
	hdata->hd.rb.scp = addr;
	hdata->hd.rb.rbp = 0;
	hdata->hd.rb.len = len;
	if (hdata->wh.dma_dir) {
		/*
		 * (at least by now) it seems, we cannot safely avoid
		 * to intercept *all* input-buffers on IP28.
		 */
		if (IP28_IP26_CODE(1 || )CACHLIGNUP(len) != len ||
		    CACHLIGNDN((unsigned)addr) != (unsigned)addr) {
			len = CACHLIGNUP(len);
			if (len <= sizeof(*hdata->hd.buf)) {
				hdata->hd.rb.rbp = hdata->hd.buf;
				addr = WD93CAC(hdata->hd.buf);
			} else {
				if ( !(addr = kmalloc(len, GFP_KERNEL)) )
					return (cmd->SCp.dma_handle = 0);
				hdata->hd.rb.rbp = addr;
			}
		}
		/* Purge cached references */
		dma_cache_inv((unsigned long) addr, len);
	}
	else /* Force flush to memory */
		dma_cache_wback((unsigned long) addr, len);
	)
	return (cmd->SCp.dma_handle = (dma_addr_t)WD93PHYS(addr));
}

static void sgiwd93_dma_unmap( Scsi_Cmnd *cmd )
{
	struct ip22_hostdata *hdata = HDATA(cmd->host);

#ifdef CONFIG_CPU_R10000
	/* R10000 on non-cachecoherent systems needs special care */
	DMA_NONCOHERENT_CODE(
	if (hdata->hd.rb.rbp)
		dma_cache_inv(WD93CAC(hdata->hd.rb.rbp),
		              CACHLIGNUP(hdata->hd.rb.len));
	else if (hdata->wh.dma_dir)
		dma_cache_inv((unsigned long)cmd->SCp.ptr,
		              cmd->SCp.this_residual);
	)
#endif
	DMA_NONCOHERENT_CODE(
	if (cmd->SCp.ptr != hdata->hd.rb.scp)
		printk(KERN_ERR "sgiwd93_dma_unmap: %s: mismatched SCp! (%p,%p)\n",
			hdata->wh.dma_dir ? "in":"out", hdata->hd.rb.scp, cmd->SCp.ptr);
	if (hdata->hd.rb.rbp) {
		if (!hdata->wh.dma_dir)
			printk(KERN_ERR "sgiwd93_dma_unmap: out: discarding stale"
				" input buffer! (%p)\n", hdata->hd.rb.scp);
		else /* if (cmd->SCp.ptr == hdata->hd.rb.scp) */
			memcpy(hdata->hd.rb.scp, hdata->hd.rb.rbp, hdata->hd.rb.len);

		if (hdata->hd.rb.rbp != hdata->hd.buf)
			kfree(hdata->hd.rb.rbp);
	}
	hdata->hd.rb.scp = 0;
	hdata->hd.rb.rbp = 0;
	hdata->hd.rb.len = 0;
	)
}

/*
 * According to Revision 1.28.2.10 we would do here instead:
 *
 * static dma_addr_t sgiwd93_dma_map( Scsi_Cmnd *cmd )
 * {  ...
 *       if (len <= sizeof(*hdata->hd.buf)) {
 *         hdata->hd.rb.rbp = addr = hdata->hd.buf;
 *         // Sigh... Screw it, to make pci_map_single() work.
 *         addr = WD93PHYS(addr) + PAGE_OFFSET;
 *       } else {
 *         ...
 *       }
 *     }
 *   })
 *   cmd->SCp.dma_handle = pci_map_single(NULL, addr, len,
 *     #if 0
 *      scsi_to_pci_dma_dir(cmd->sc_data_direction));
 *     #else
 *      hdata->wh.dma_dir ? PCI_DMA_FROMDEVICE:PCI_DMA_TODEVICE);
 *     #endif
 *   ret...
 * }
 * static void sgiwd93_dma_unmap( Scsi_Cmnd *cmd )
 * { ...
 *   #if defined(CONFIG_SGI_IP28)
 *     ...
 *   #endif
 *   pci_unmap_single(NULL, cmd->SCp.dma_handle, len,
 *             hdata->wh.dma_dir ? PCI_DMA_FROMDEVICE:PCI_DMA_TODEVICE);
 *   DMA_NONCOHERENT_CODE(
 *     ...
 *   )
 * }
 *
 * But we omit pci_map_single() because:
 * 1) It will dereference the NULL argument...
 * 2) It demands a corresponding pci_unmap_single(), which in turn will:
 * 1) Dereference the NULL argument...
 * 2) Overwrite input (!) buffers with dma_cache_wback_inv() (=8-{)
 *    This is just, what we must inhibit (e.g. by a preceding
 *    dma_cache_inv() ;-)
 *    (at least with R10k on non-cache-coherent systems, on other machines
 *     we might be lucky and the cache will be clean, making wback a nop)
 *
 * (Doing always dma_cache_wback_inv() in pci_map_single() is alright,
 *  but dma_cache_inv() would be sufficient for input buffers)
 *
 * dma_map_single()/dma_unmap_single() in 2.6.x do the right thing.
 *
 * Btw: we cannot use 'cmd->sc_data_direction' here, since we are
 * called for an explicit dma-direction, and it is mere *coincidence*,
 * if 'sc_data_direction' is consistent with this direction !
 * (tracing shows, 'sc_...' isn't well defined at all...)
 */

#define HOSTS_C

#include "sgiwd93.h"

static Scsi_Host_Template driver_template = SGIWD93_SCSI;

#include "scsi_module.c"

int sgiwd93_release(struct Scsi_Host *instance)
{
#ifdef MODULE
	struct ip22_hostdata *hdata = HDATA(instance);
    int irq = 0;

	if (sgiwd93_host && sgiwd93_host == instance)
		irq = SGI_WD93_0_IRQ;
	else if (sgiwd93_host1 && sgiwd93_host1 == instance)
		irq = SGI_WD93_1_IRQ;

	free_irq(irq, sgiwd93_intr);
	pci_free_consistent(NULL, DMA_NONCOHERENT_CODE(2*)PAGE_SIZE,
	                    hdata->hd.cpu, hdata->hd.dma);
	wd33c93_release();
#endif
	return 1;
}
/*
 * Revision 1.28.2.10, Sat Oct 11 07:42:43 2003
 * May/Dec 2004 pf - xkphys, cache-alignment, IP28/IP26
 */
